#include "tsp.h"
#include "hrtimer_x86.h"
#include <sys/time.h>
#include <mpi.h>
#include <math.h>

void err_sys(const char *msg)
{
	printf("%s\n",msg);
 	exit(1);
}

int main( int argc, char* argv[])
{
	int ncities;
  	int nthreads;
   	int rank;
   	int i;
   	List result;
   	Arg* arg = NULL;
	double before, after;

   	MPI_Init(&argc, &argv);
   	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   	MPI_Comm_size(MPI_COMM_WORLD, &nthreads);

   	if (rank == 0) 
	{
      		if (fscanf(stdin, "%d\n", &ncities) != 1) {
			err_sys("can not read from stdin\n");
		}
      		printf("Use 1 master process and %d slave processes for %d cities.\n", nthreads-1, ncities);
      		MPI_Bcast(&ncities, 1, MPI_INT, 0, MPI_COMM_WORLD);
		before = gethrtime_x86();
   	}

   	initialize_list(&result);

   	if (rank != 0)
	{
      		arg = (Arg *)malloc(sizeof(Arg));
      		arg->nsubcities = 0;
      		initialize_list(&arg->city_list);
   	}
 
   	// assign sub cities for each threads. 
   	subcities(arg, ncities, nthreads);

   	if (rank != 0)
	{
      		TSP_thread(arg);
   	}
   
   	mergeCities(&result, arg, nthreads);
   
   	if (rank == 0)
	{
		//sleep(3000);
		after = gethrtime_x86();
		printf("running time of MPI solution is %f sec\n",after-before);
      		output_file(&result);
	}

   	MPI_Finalize();
}


int sqrt_my(int num) 
{
   	int op = num;
   	int res = 0;
   	int one = 1 << 14; // The second-to-top bit is set: 1L<<30 for long
                      // "one" starts at the highest power of four <= the argument.
   	while (one > op)
      		one >>= 2;
      
      	while (one != 0) 
	{
         	if (op >= res + one) 
		{
            		op -= res + one;
            		res += 2 * one;
         	}
         	res >>= 1;
         	one >>= 2;
      	}
   	return res;
}

void subcities(Arg* arg, int ncities, int nthreads) 
{
   	int regions = sqrt_my(nthreads - 1);
   	unsigned long  block_range = RANGE / regions;
   	int block_X, block_Y;
	int flag,tag;
   	int i;
	int rank;
   	int thread;
   	City city;
        MPI_Status status;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	if (rank == 0)
	{

   		for( i = 0; i<ncities ; i++ )
		{
      			if (fscanf(stdin, "%lu %lu\n", &city.X, &city.Y) != 2) {
				err_sys("can not read from stdin\n");
			}
      			city.X = city.X % RANGE, city.Y = city.Y % RANGE;
      			block_X = (int) city.X / block_range, block_Y = (int) city.Y / block_range;

      			if (block_Y % 2 == 1)
			{
         			block_X = (regions-1) - block_X;
      			}
      
      			thread = regions * block_Y + block_X;
                        thread++;

			flag = sizeof(City);
         		MPI_Send(&flag, 1, MPI_INT, thread, 100, MPI_COMM_WORLD);
         		MPI_Send((char*)&city, sizeof(City), MPI_CHAR, thread, 100, MPI_COMM_WORLD);
   		}
		for (i = 1; i < nthreads ; i++)
		{
         		flag = 0;
         		MPI_Send(&flag, 1, MPI_INT, i, 100, MPI_COMM_WORLD);
      		}
	}

	else
	{
		while(1)
		{
         		MPI_Recv(&flag, 1, MPI_INT, 0, 100, MPI_COMM_WORLD, &status);
         		if (flag != 0) 
			{
            			MPI_Recv(&city, sizeof(City), MPI_CHAR, 0, 100, MPI_COMM_WORLD, &status);
            			add_to_tail( &(arg->city_list), &city, sizeof(City) );
            			arg->nsubcities += 1;
         		} 
			else 
            			break;
         	}

	}
}


int same_side(Node* V1, Node* V2, Node* U1, Node* U2) 
{
   	// if U1 and U1 are at the same side of line y=mx+b (V1 -- V2)
   	// return 1; else 0;
   	double m, b, c1, c2;
   	double u1,v1,u2;
   	double y2, y1, x2, x1;
   	int res;
  
   	// vertical line, m = infinite, must calculated otherwise
   	if (( ((City*)V2->data)->X - ((City*)V1->data)->X ) == 0)
	{
      		u1 = ((City*)U1->data)->X;
      		v1 = ((City*)V1->data)->X;
      		u2 = ((City*)U2->data)->X;
      
      		if ( u1 - v1 >= 0 && u2 - v1 >= 0    ||
		     u1 - v1 <= 0 && u2 - v1 <= 0 )
		{
         		return 1;
      		} 
		else 
         		return 0;
   	}

   	// m = (y2-y1)/(x2-x1)
   	y2 = ((City*)V2->data)->Y;
   	y1 = ((City*)V1->data)->Y;
   	x2 = ((City*)V2->data)->X;
   	x1 = ((City*)V1->data)->X;
   	m = (y2-y1)/(x2-x1);
   
   	// b = y-mx
   	b = ((City*)V1->data)->Y - m * ((City*)V1->data)->X;
   
   	// test same side use c = y-mx-b 'c > 0' or 'c < 0'
   	c1 = ((City*)U1->data)->Y - m * ((City*)U1->data)->X - b;
   	c2 = ((City*)U2->data)->Y - m * ((City*)U2->data)->X - b;
   
   	if ( (c1 >= 0 && c2 >= 0) || (c1 <= 0 && c2 <= 0) )
	{
      		res = 1;
   	} 
	else 
	{
      		res = 0;
   	}
   	return res;
}


int cross(Node* U1, Node* U2, Node* V1, Node* V2)
{
   	int res;
   	// if U1 and U1 are at the same side of line (V1 -- V2)
   	if ( same_side (V1, V2, U1, U2) ) 
	{
      		res = 0;
   	// if V1 and V2 are at the same side of line (U1 -- U2)
   	} 
	else if ( same_side (U1, U2, V1, V2) )
	{
      		res = 0;
   	} 
	else 
	{
     		res = 1;
   	}
   	return res;
}


void *TSP_thread(void * arg) 
{
   	unsigned long ncities = ((Arg*)arg)->nsubcities;
   	List cities = ((Arg*)arg)->city_list;
   	unsigned long times =  10 * ncities * ncities;
   	unsigned long i;
   	unsigned long node_1, node_2, tmp;

   	Node** index = (Node**)malloc(sizeof(Node*) * ncities);
   	Node* cur = cities.head;
   	// build a index array
   	for (i=0; i<ncities; i++)
	{
      		index[i] = cur;
      		cur = cur->next;
   	} 

   	srand((int)time(0));
   	for (i=0 ; i<times ; i++) 
	{
      		node_1 = rand() % ncities;
      		node_2 = rand() % ncities;
      		// let node_1 be the small one
      		if (node_1 > node_2)
		{
	      		tmp = node_1;
	      		node_1 = node_2;
	      		node_2 = tmp;
      		}
      		// don't do that on the head and tail
      		if(node_1 == 0)
	      		continue;
      		if(node_2 == (ncities -1))
	      		continue;
      		// done if the same node
      		if(node_1 == node_2)
	      		continue;
   
      		// do i need to do path exchange?
      		if (cross(index[node_1-1],index[node_1],index[node_2],index[node_2+1])) 
		{
         		// 12 34567 8 -> 12 76543 8
         		index[node_1-1]->next = index[node_2];
         		tmp = node_1 + 1;	 
	 		while( tmp <= node_2) 
			{
            			index[tmp]->next = index[tmp-1];
	    			tmp ++;
	 		}
	 		index[node_1]->next = index[node_2+1];
	 		// update index
	 		tmp = node_1;
	 		cur = index[node_1-1];
	 		while( cur != index[node_2+1]) 
			{
            			index[tmp] = cur->next;
	    			cur = cur->next;
	    			tmp ++;
         		}
      		}
   	}
   	return;
}

void mergeCities(List* result, Arg * arg, int nthreads) 
{
   	int i;
	int rank;
	int flag;

	Node* index;
        City city;
	int subcities;
	MPI_Status status;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	if (rank == 0)
	{
		for(i = 1; i < nthreads; i++)
		{
         		while(1)
			{
            			MPI_Recv(&flag, 1, MPI_INT, i, 100, MPI_COMM_WORLD, &status);
            			if (flag == 0)
					break;
            			MPI_Recv((char*)&city, sizeof(City), MPI_CHAR, i, 100, MPI_COMM_WORLD, &status);
            			add_to_tail( result, &city, sizeof(City) );
         		}
      		}
	}

	else
	{
		if (arg->nsubcities != 0) 
		{
         		index = arg->city_list.head;
         		subcities = arg->nsubcities;
         		for ( i = 0; i < subcities; i++)
			{
            			flag = sizeof(City);
            			MPI_Send(&flag, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
                      		MPI_Send((char*)index->data, sizeof(City), MPI_CHAR, 0, 100, MPI_COMM_WORLD);
            			index = index->next;
         		}
         		flag = 0;
         		MPI_Send(&flag, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
      		}
		else 
		{
         		flag = 0;
         		MPI_Send(&flag, 1, MPI_INT, 0, 100, MPI_COMM_WORLD);
      		}
   	}
}


void output_file(List * list) 
{
   	FILE * fp;
   	Node * from;
   	Node * to;
   	City * city_from;
   	City * city_to;
	double length = 0.0;
	unsigned long x1, x2, y1, y2;

   	if( (fp=fopen("result", "w")) == NULL) 
	{
      		err_sys("can not open file to write result\n");
   	}

   	fprintf(fp, "TSP problem\n");
   	fprintf(stdout, "TSP problem\n");
   	fprintf(fp, "%d %d %d %d\n", 0 , RANGE, 0, RANGE);
   	fprintf(stdout, "%d %d %d %d\n", 0 , RANGE, 0, RANGE);

   	from = list->head;
   	to = list->head->next;
   	while (from->next != NULL) 
	{
      		city_from = (City*)from->data;
      		city_to = (City*)to->data;
		
		x1 = city_from->X; 
		y1 = city_from->Y; 
		x2 = city_to->X;
		y2 = city_to->Y; 
      		fprintf(fp, "+ %lu %lu %lu %lu\n", city_from->X, city_from->Y, city_to->X, city_to->Y);
      		fprintf(stdout, "+ %lu %lu %lu %lu\n", city_from->X, city_from->Y, city_to->X, city_to->Y);
      		fprintf(fp, "time: 1\n");
      		fprintf(stdout, "time: 1\n");
		length = length + sqrt((double)((x2-x1)*(x2-x1))+(double)((y2-y1)*(y2-y1)));
      		from = to;
      		to = to->next;
   	}
   	city_from = (City*)from->data;
   	city_to = (City*)list->head->data;
   	fprintf(fp, "+ %lu %lu %lu %lu\n", city_from->X, city_from->Y, city_to->X, city_to->Y);
   	fprintf(stdout, "+ %lu %lu %lu %lu\n", city_from->X, city_from->Y, city_to->X, city_to->Y);
   	fprintf(fp, "time: 1\n");
   	fprintf(stdout, "time: 1\n");
	printf("Total length is %f\n", length);
   
   	fclose(fp);
}
